#ifndef ENTITY_H_INCLUDED
#define ENTITY_H_INCLUDED

#include <iostream>
#include <fstream>      /* pour ecrire dans un fichier */
#include <cmath>        /* outils mathematiques */
#include <algorithm>
#include <stdlib.h>     /* srand, rand */
#include <vector>
#include "rateMove.h"
#include "Geometry.h"
#include <assert.h>

class Entity
{
    private:
        // traits phenotypiques de l'individu integrer dans un vector (il s'agit de tous les traits
        // qui ne sont pas modifies au cours de la vie d'un individu)
        std::vector<double> m_traitPhenotype;
        // position de l'individu, c'est-a-dire un trait modifie pendant sa vie, pour l'instant dim1
        double m_traitPosition;
        // dernier temps où la position de l'individu a ete calculee
        double m_timePosition;
        //l' attribut suivant est commun a tous les individus de la population
        Geometry m_geometrySpace;
        // retient tous les parametres lies a la population ou a l'individu
        // le premier correspond au numero de l'espece a laquelle l'individu appartient
        // ex: 1.range d'interaction competitive, 2.variance du trait mutant, 3.coeffDiffusion
        std::vector<double> m_parameters;
        //taux de mutation associe a l'espece de l'individu
        double m_probaMutation;

        void updatePosition(double const& timeNow);

    public:

        //voir commentaires dans le .cpp
        Entity(std::vector<double> const& traitPhenotype, double const& traitPosition,
               double const& timeNow, Geometry const& geometrySpace,
               std::vector<double> const& parameters, double const& probaMutation);
        ~Entity();
        //change la position de l'individu, utiliser uniquement pour l'initialisation
        void changePosition(double const& newPosition);
        Geometry getGeometry() const;
        double getProbaMutation() const;
        double getFirstTraitPhenotype() const;
        double coefDiffusionEntity() const;
        double birthRateEntity(double const& timeNow);
        double positiveInteraction(double const& timeNow, Entity partner);
        double deathRateEntity(double const& timeNow);
        double competition(double const& timeNow, Entity competitor);
        double competitionMut(double const& timeNow, Entity competitor);
        Entity mutant(double const& timeNow);
        Entity mutantShiftBirth(double const& timeNow);
        void writeEntity(std::string nameFileOut, double const& timeNow);
        void defineUplet(double const& timeNow, double const& sizeIntervalPhenotype,
                         double const& sizeIntervalPosition, int &index1, int &index2);
        int getType() const;
        bool testPosition(double const& timeNow, double const& maxPosition); 
};


#endif // ENTITY_H_INCLUDED
