#include "Population.h"

using namespace std;

//faire une classe population! contenant un tableau dynamique d'individus

//constructeur de base
Population::Population()
{
    vector<vector<Entity> > newtable;
    m_tableEntities = newtable;
}


//constructeur
Population::Population(Entity const& firstEntity)
{
    int Type(firstEntity.getType());
    if (Type!=0) 
    {
        cout << "first Entity must have type 0" << endl;
        assert(false);
    }
    vector<Entity> newPopulation(1,firstEntity);
    vector<vector<Entity> > newtable(1,newPopulation);
    m_tableEntities = newtable;
}


// destructeur
Population::~Population()
{
    //for ( int k(m_tableEntities.size() - 1), k > -1 , k-- )
    //    {
    //    delete m_tableEntities[k];        //supprimer l'entite pointee
    //    m_tableEntities[k] = 0;           // mettre le pointeur a 0 pour eviter tout probleme!
    //    m_tableEntities.pop_back();       // supprimer la derniere case du tableau
    //    }
}


//renvoie le nombre d'Entities presentes
int Population::nbrEntities() const
{
    int number(0);
    for (int k(0); k < nbrTypes(); k++)
    {
        number += nbrEntitiesInType(k);
    }
    return number;
}


int Population::nbrEntitiesInType(int const& typeEntity) const
{
    return m_tableEntities[typeEntity].size();
}

int Population::nbrTypes() const
{
    return m_tableEntities.size();
}


//verifie si il existe encore une entite en vie
bool Population::isAlive() const
{
    if ( nbrEntities() == 0 )
         return false;
    else
         return true;
}


//calcule le taux de competition entre deux Entities
//attention, modifie les deux Entities concernees: modification des positions
double Population::killRateCompetition(double const& timeNow, int const& indexEntity1,
                                       int const& typeEntity1, int const& indexEntity2,
                                       int const& typeEntity2)
{
    
    return m_tableEntities[typeEntity1][indexEntity1].competition(timeNow, m_tableEntities[typeEntity2][indexEntity2]);
}

//calcule le taux de competition pour l'acces aux mutualistes entre deux Entities
//attention, modifie les deux Entities concernees: modification des positions
double Population::killRateCompetitionMut(double const& timeNow, int const& indexEntity1,
                                          int const& typeEntity1, int const& indexEntity2,
                                          int const& typeEntity2)
{
    return m_tableEntities[typeEntity1][indexEntity1].competitionMut(timeNow, m_tableEntities[typeEntity2][indexEntity2]);
}


//calcule le taux de mort de l'Entity concernee
//attention, modifie l'Entity concernee: mis a jour de la position
double Population::killRateNatural(double const& timeNow, int const& indexEntity,
                                   int const& typeEntity)
{
    return m_tableEntities[typeEntity][indexEntity].deathRateEntity(timeNow);
}



//tuer l'individu nKilled et le remplacer par le dernier du tableau
void Population::killEntity(int const& nKilled, int const& typeKilled)
{
    if ( !isAlive() )
        {
        cout << "all entities are dead";
        exit(0);
        }
    else
        {
        //m_tableEntities.erase(m_tableEntities.begin() + nKilled);
        //la technique ci-dessus est tres lente!
        int lastIndex(m_tableEntities[typeKilled].size() - 1);
        m_tableEntities[typeKilled][nKilled] = m_tableEntities[typeKilled][lastIndex];
        m_tableEntities[typeKilled].pop_back();
        };
}



//calcule le taux de naissance de l'Entity concernee
//attention, modifie l'Entity concernee: mis a jour de la position
double Population::birthRateNatural(double const& timeNow, int const& indexEntity,
                                    int const& typeEntity)
{
    return m_tableEntities[typeEntity][indexEntity].birthRateEntity(timeNow);
}


//calcule le taux d'interaction positive entre deux Entities
//attention, modifie les deux Entities concernees: modification des positions
double Population::birthRateInteraction(double const& timeNow, int const& indexEntity1,
                                        int const& typeEntity1, int const& indexEntity2,
                                        int const& typeEntity2)
{
    return m_tableEntities[typeEntity1][indexEntity1].positiveInteraction(timeNow, m_tableEntities[typeEntity2][indexEntity2]);
}


//verifie si l'entite a ajouter fait partie d'une espece connue ou non (quitte a l'ajouter)
bool Population::knownSpecies(int const& typeEntity) const
{
    return (nbrTypes()>typeEntity);
}


//ajouter l'Entity oneAdded a la population
void Population::addEntity(Entity const& oneAdded)
{
    int typeEntity(oneAdded.getType());
    if ( knownSpecies(typeEntity) )
    {
        m_tableEntities[typeEntity].push_back(oneAdded);
    }
    else
    {
    if ( typeEntity == nbrTypes() )
    {
        vector<Entity> newPopulation(1,oneAdded);
        m_tableEntities.push_back(newPopulation);
    }
    else
    {
        cout << "impossible to add this species type" << endl;
        assert(false);
    }
    }
}


//supprime tous les individus en vie et retourne une population vide
void Population::empty()
{
    vector<vector<Entity> > newtable;
    m_tableEntities = newtable;
}    



// Une etape: mort par competition, mort naturelle, ou naissance (avec ou sans clone)
void Population::chooseEvent(double const& timeNow, double const& birthRate, double const& deathRate,
                         double const& interRate, int const& typePartner, double const& compIntraRate,
                         const double &thetaInitial, int const& indexEndangered,
                         int const& typeEndangered, bool &isDead, bool &isBirth)
{
    //theta est le tirage uniforme
    double theta(thetaInitial);

    //controle s'il y a un evenement et de quel type, les evenements sont realises a la fin
    isDead = false;
    isBirth = false;

    if (theta < birthRate)
    {
        //cas de naissance sans mutualiste potentielle
        //calcul du taux reel de naissance naturelle a partir des positions actualisees
        double birthReal( birthRateNatural(timeNow, indexEndangered, typeEndangered) * birthRate );
            //verification si il y a naissance effectivement
        if (theta  < birthReal )  isBirth = true;
    }
    else
    {
    theta -= birthRate;
    if (theta < deathRate)
    {
        //cas de mort naturelle potentielle
        //calcul du taux reel de mort naturelle a partir des positions actualisees
        double deathReal( killRateNatural(timeNow, indexEndangered, typeEndangered) * deathRate );
            //verification si il y a mort effectivement
        if (theta  < deathReal )  isDead = true;
    }
    else
    {
    theta -= deathRate;
    if (theta < interRate*nbrEntitiesInType(typePartner))
    {
        //cas de naissance potentielle
        //indice du partenaire d'interaction positive
        int indexPartner((int)(theta / interRate));
        //calcul du taux reel de naissance a partir des positions actualisees
        double posInterReal( birthRateInteraction(timeNow, indexEndangered, typeEndangered, 
                             indexPartner, typePartner) * interRate );
            //verification si il y a mort effectivement
        if (abs(theta - indexPartner*interRate) < posInterReal )  isBirth = true;

    }

    else
    {
    theta -= interRate*nbrEntitiesInType(typePartner);
    if (theta < compIntraRate*nbrEntitiesInType(typeEndangered))
    {
        //cas "competition"
        //indice du partenaire de competition
        int indexCompetitor((int)(theta / compIntraRate));
        //calcul du taux reel de competition a partir des positions actualisees
        double compReal( killRateCompetition(timeNow, indexEndangered, typeEndangered,
                         indexCompetitor, typeEndangered) * compIntraRate );
        //verification si il y a mort effectivement
        if (abs(theta - indexCompetitor*compIntraRate) < compReal )  isDead = true;
    }
    }
    }
    }
}

void Population::oneStep(double const& timeNow, double const& birthRate, double const& deathRate,
                         double const& interRate, int const& typePartner, double const& compIntraRate,
                         const double &thetaInitial, int const& indexEndangered,
                         int const& typeEndangered)
{
    //choix de l'evenement de mort ou de naissance
    bool isDead, isBirth;
    chooseEvent(timeNow, birthRate, deathRate, interRate, typePartner, compIntraRate,
                thetaInitial, indexEndangered, typeEndangered, isDead, isBirth);

    //realisation des evenements mort/naissance
    if (isDead)
    {
        killEntity(indexEndangered, typeEndangered);
    }
    if (isBirth)
    {
        Entity newEntity(m_tableEntities[typeEndangered][indexEndangered].mutantShiftBirth(timeNow));
        //mutation?
        double rvUniform((double)rand()/((double)RAND_MAX+1.0));
        double mutationProbability(m_tableEntities[typeEndangered][indexEndangered].getProbaMutation());
        //cas de mutation
        if (rvUniform < mutationProbability) newEntity = newEntity.mutant(timeNow);
        addEntity(newEntity);
    }
}


// Une etape: mort par competition, mort naturelle, ou naissance (avec ou sans clone)
void Population::oneStep_isMutation(double const& timeNow, double const& birthRate, double const& deathRate,
                         double const& interRate, int const& typePartner, double const& compIntraRate,
                         const double &thetaInitial, int const& indexEndangered,
                         int const& typeEndangered, bool &isMutation)
{
    //choix de l'evenement de mort ou de naissance
    bool isDead, isBirth;
    chooseEvent(timeNow, birthRate, deathRate, interRate, typePartner, compIntraRate,
                thetaInitial, indexEndangered, typeEndangered, isDead, isBirth);

    //realisation des evenements mort/naissance
    if (isDead)
    {
        killEntity(indexEndangered, typeEndangered);
    }
    if (isBirth)
    {
        Entity newEntity(m_tableEntities[typeEndangered][indexEndangered].mutantShiftBirth(timeNow));
        //mutation?
        double rvUniform((double)rand()/((double)RAND_MAX+1.0));
        double mutationProbability(m_tableEntities[typeEndangered][indexEndangered].getProbaMutation());
        //cas de mutation
        if (rvUniform < mutationProbability) 
        {
            newEntity = newEntity.mutant(timeNow);
            isMutation = true;
        }
        else
        {
            isMutation = false;
        }
        addEntity(newEntity);
    }
}




void Population::oneStepTwoSpecies(double& timeNow, double const& maxBirthRatebees, double const& maxDeathRatebees, 
                               double const& maxInterRatebees, double const& maxCompRatebees, 
                               double const& maxBirthRateplants, double const& maxDeathRateplants, 
                               double const& maxInterRateplants, double const& maxCompRateplants)
{
        //calcul du taux max d'evenement
        // type 0=abeille; type 1=plantes
        //taux abeille
        double beeMaxRate(maxBirthRatebees + maxDeathRatebees + maxInterRatebees * nbrEntitiesInType(1) +
                maxCompRatebees * nbrEntitiesInType(0));
        //taux plante
        double plantMaxRate(maxBirthRateplants + maxDeathRateplants + maxInterRateplants * nbrEntitiesInType(0) +
                maxCompRateplants * nbrEntitiesInType(1));
        //taux maximal
        double maxRate(beeMaxRate * nbrEntitiesInType(0) + plantMaxRate * nbrEntitiesInType(1));

        //choix du temps de l'evenement
        double T( -log( 1.0 - ((double)rand()+1.0)/((double)RAND_MAX+2.0) ) );      //variable aleatoire expo(1)
        timeNow += T / maxRate;

        //choix de l'indexEndangered+choix de l'evenement en utilisant la function oneStep
        int indexEndangered;
        double U( (double)rand()/((double)RAND_MAX+1.0) * maxRate );
        double V(0);
        if (U < beeMaxRate * nbrEntitiesInType(0))
        {
            V = (double)rand()/((double)RAND_MAX+1.0) * beeMaxRate;
            indexEndangered = rand() % nbrEntitiesInType(0);
            oneStep(timeNow, maxBirthRatebees, maxDeathRatebees, maxInterRatebees, 1, 
                    maxCompRatebees, V, indexEndangered, 0);
        }
        else
        {
            V = (double)rand()/((double)RAND_MAX+1.0) * plantMaxRate;
            indexEndangered = rand() % nbrEntitiesInType(1);
            oneStep(timeNow, maxBirthRateplants, maxDeathRateplants, maxInterRateplants, 0,
                    maxCompRateplants, V, indexEndangered, 1);
        }

}


void Population::oneStepOneSpecies(double& timeNow, double const& maxBirthRatebees, double const& maxDeathRatebees, 
                                   double const& maxCompRatebees, bool &isMutation)
{
        //calcul du taux max d'evenement
        // type 0=abeille; type 1=plantes
        //taux abeille
        double beeMaxRate(maxBirthRatebees + maxDeathRatebees +
                maxCompRatebees * nbrEntitiesInType(0));
        //taux maximal
        double maxRate(beeMaxRate * nbrEntitiesInType(0));

        //choix du temps de l'evenement
        double T( -log( 1.0 - ((double)rand()+1.0)/((double)RAND_MAX+2.0) ) );      //variable aleatoire expo(1)
        timeNow += T / maxRate;

        //choix de l'indexEndangered+choix de l'evenement en utilisant la function oneStep
        int indexEndangered;
        double V(0);
        V = (double)rand()/((double)RAND_MAX+1.0) * beeMaxRate;
        indexEndangered = rand() % nbrEntitiesInType(0);
        oneStep_isMutation(timeNow, maxBirthRatebees, maxDeathRatebees, 0, 0, 
                    maxCompRatebees, V, indexEndangered, 0, isMutation);
}



bool Population::update_test(double const& timeNow, double const& testedPosition)
{
    bool test(false);
    for (int k(0); k < abs(nbrTypes()); k++)
    {
        for (int j(0); j < abs(nbrEntitiesInType(k)); j++)
        {
            if (m_tableEntities[k][j].testPosition(timeNow,testedPosition))
                test = true;
        }
    }
    return test;
}



//Ecrit dans un fichier nameFileOut, tous les entities de la population,
//une ligne par Entity
void Population::writePopulation(string nameFileOut, double const& timeNow)
{
    for (int k(0); k < abs(nbrTypes()) ; k++)
    {
        for (int j(0); j < abs(nbrEntitiesInType(k)); j++)
        {
            m_tableEntities[k][j].writeEntity(nameFileOut, timeNow);
        }
    }
}


//Ecrit dans un fichier nameFileOut, tous les traits presents dans la population 
//et le nombre d'individus possedant le trait: type nbrInd type nbrInd...
//le tout sur une ligne 
//ATTENTION: cette fonction ne met pas a jour les positions des individus pour gain de temps CPU
void Population::writeTraits(string nameFileOut, double const& timeNow)
{
    double traitPhenotype(m_tableEntities[0][0].getFirstTraitPhenotype());
    vector<double> traits(1,traitPhenotype);
    vector<int> nbrInds(1,0);
    bool in(false);
    int sizeTraits(1);

    for (int k(0); k < abs(nbrTypes()); k++)
    {
        for (int j(0); j < abs(nbrEntitiesInType(k)); j++)
        {
            traitPhenotype = m_tableEntities[k][j].getFirstTraitPhenotype();
            in = false;
            for (int l(0); l < sizeTraits; l++) 
            {
            if (traits[l] == traitPhenotype)
            {
                nbrInds[l] +=1;
                in = true;
            }
            }
            if (!in)
            {
                traits.push_back(traitPhenotype);
                nbrInds.push_back(1);
                ++sizeTraits;
            }
        }
    }

    ofstream flux(nameFileOut.c_str(), ios::app);
    flux << timeNow << " ";    

    for (int m(0); m < sizeTraits; m++)
    {
        flux << traits[m] << " " << nbrInds[m] << " ";
    }

    flux << endl;
}


